/*
* nuage3D.c
* 
* Copyright (C) 2016, Tristan Dagobert, CMLA, École Normale Supérieure de Cachan.
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or 
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "nuage3D.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
/*===========================================================================*/
#define VERSION "1.1"

/*===========================================================================*/
void
reading_image_dimensions (char *filename_contributions, int xmax[1],
			  int ymax[1])
{
  FILE *file_contributions = NULL;
  int nlig, ncol;
  int taille_int = sizeof (int);
  int ret;

  file_contributions = fopen (filename_contributions, "rb");
  if (file_contributions == NULL)
    {
      perror ("reading_image_dimensions()\n");
      exit (EXIT_FAILURE);
    }

  /* total size of image */
  ret = fread (&nlig, taille_int, 1, file_contributions);
  if (ret != 1)
    {
      printf ("read()\n");
      exit (EXIT_FAILURE);
    }

  ret = fread (&ncol, taille_int, 1, file_contributions);
  if (ret != 1)
    {
      printf ("read()\n");
      exit (EXIT_FAILURE);
    }

  xmax[0] = ncol - 1;
  ymax[0] = nlig - 1;

  fclose (file_contributions);
  printf ("[INFO] images dimensions : %d col. x %d lines\n", xmax[0] + 1,
	  ymax[0] + 1);
  return;
}

/*===========================================================================*/
contributions_t **
allocating_contributions (int nlig, int ncol)
{
  contributions_t **contributions = NULL;
  int i, j;

  contributions =
    (contributions_t **) malloc (nlig * sizeof (contributions_t *));
  if (contributions == NULL)
    {
      perror ("allocating_contributions()\n");
      exit (EXIT_FAILURE);
    }
  for (i = 0; i < nlig; i++)
    {
      contributions[i] =
	(contributions_t *) malloc (ncol * sizeof (contributions_t));
      if (contributions[i] == NULL)
	{
	  perror ("allocating_contributions()\n");
	  exit (EXIT_FAILURE);
	}
      for (j = 0; j < ncol; j++)
	{

	  contributions[i][j].nb_contribs = 0;
	  contributions[i][j].xpos = 0;
	  contributions[i][j].ypos = 0;
	  contributions[i][j].x = NULL;
	  contributions[i][j].y = NULL;
	  contributions[i][j].c = NULL;
	}
    }
  return contributions;
}

/*===========================================================================*/
contributions_t **
loading_contributions (char *filename_contributions,
		       int xmin, int ymin, int xmax, int ymax, int *ncbs)
{
  int taille_int = sizeof (int);
  int taille_float = sizeof (float);
  int taille_3float = 3 * taille_float;

  int x, y, xpos, ypos;
  int n, nb_contributions = 0;
  long saut;

  int nlig, ncol;
  FILE *file_contributions = NULL;

  contributions_t **contributions;

  int ret;
  int ypos_et_xpos[2];
  float triplet[3];

  int i, j, k;

  file_contributions = fopen (filename_contributions, "rb");
  if (file_contributions == NULL)
    {
      perror ("loading_contributions\n");
      exit (EXIT_FAILURE);
    }

  /* computing the image size */
  ret = fread (&nlig, taille_int, 1, file_contributions);
  if (ret != 1)
    {
      perror ("loading_contributions\n");
      exit (EXIT_FAILURE);
    }

  ret = fread (&ncol, taille_int, 1, file_contributions);
  if (ret != 1)
    {
      perror ("loading_contributions\n");
      exit (EXIT_FAILURE);
    }

  ret = fread (ncbs, taille_int, 1, file_contributions);
  if (ret != 1)
    {
      perror ("loading_contributions\n");
      exit (EXIT_FAILURE);
    }

  /* building the matrix of contributions */
  contributions = allocating_contributions (ymax - ymin + 1, xmax - xmin + 1);
  if (contributions == NULL)
    {
      perror ("loading_contributions\n");
      exit (EXIT_FAILURE);
    }

  for (y = ymin, i = 0; y <= ymax; y++, i++)
    {
      for (x = xmin, j = 0; x <= xmax; x++, j++)
	{
	  /* positionning on the current pixel */
	  ret = fread (&ypos_et_xpos, 2 * taille_int, 1, file_contributions);
	  if (ret != 1)
	    {
	      if (feof (file_contributions))
		continue;
	      printf ("loading_contributions() lecture x y\n");
	      exit (EXIT_FAILURE);
	    }
	  ypos = ypos_et_xpos[0];
	  xpos = ypos_et_xpos[1];

	  /* positionning on the ad hoc pixel */
	  while (ypos != y || xpos != x)
	    {
	      /* positionning on the number of contributions */
	      ret =
		fread (&nb_contributions, taille_int, 1, file_contributions);
	      if (ret != 1)
		{
		  if (feof (file_contributions))
		    {
		      printf ("abnormal end of file\n");
		    }
		  else
		    {
		      printf ("reading problem\n");
		    }
		  exit (EXIT_FAILURE);
		}

	      /* leapfrogging... */
	      saut = (long) nb_contributions *3 * taille_float;

	      if (saut > 0)
		{
		  ret = fseek (file_contributions, saut, SEEK_CUR);
		  if (ret != 0)
		    {
		      printf ("loading_contributions() abnormal position\n");
		    }
		}
	      /* if nature is well done, then we are on the next pixel */
	      ret =
		fread (&ypos_et_xpos, 2 * taille_int, 1, file_contributions);
	      if (ret != 1)
		{
		  if (feof (file_contributions))
		    {
		      printf ("loading_contributions() abnormal ending\n");
		      exit (EXIT_FAILURE);
		    }
		}
	      ypos = ypos_et_xpos[0];
	      xpos = ypos_et_xpos[1];
	    }

	  /* we are now on the ad hoc pixel : we are going to read the set
	   * of its contributions ... */
	  ret = fread (&nb_contributions, taille_int, 1, file_contributions);
	  if (nb_contributions == 0)
	    {
	      contributions[i][j].xpos = (short int) xpos;
	      contributions[i][j].ypos = (short int) ypos;

	      continue;
	    }

	  /* family allowances... */
	  contributions[i][j].xpos = (short int) xpos;
	  contributions[i][j].ypos = (short int) ypos;
	  contributions[i][j].x =
	    (float *) calloc (nb_contributions, taille_float);
	  contributions[i][j].y =
	    (float *) calloc (nb_contributions, taille_float);
	  contributions[i][j].c =
	    (float *) calloc (nb_contributions, taille_float);

	  if (contributions[i][j].x == NULL
	      || contributions[i][j].y == NULL
	      || contributions[i][j].c == NULL)
	    {
	      perror ("loading_contributions() : allocations\n");
	      exit (EXIT_FAILURE);
	    }

	  n = 1;
	  k = 0;
	  while (n <= nb_contributions)
	    {
	      /* reading coordinates */
	      ret = fread (triplet, taille_3float, 1, file_contributions);
	      if (ret != 1)
		{
		  perror ("loading_contributions() : reading triplet\n");
		  exit (EXIT_FAILURE);
		}
	      /* normality test and crisis management... */
	      ret = fpclassify (triplet[2]);
	      switch (ret)
		{
		case FP_NAN:
		  {
		    printf ("NaN contribution for i=%d j=%d\n", i, j);
		    break;
		  }
		case FP_INFINITE:
		  {
		    triplet[2] = FLT_MAX;
		    contributions[i][j].x[k] = triplet[0];
		    contributions[i][j].y[k] = triplet[1];
		    contributions[i][j].c[k] = FLT_MAX;
		    k++;
		    break;
		  }
		case FP_SUBNORMAL:
		  {
		    triplet[2] = FLT_MIN;
		    contributions[i][j].x[k] = triplet[0];
		    contributions[i][j].y[k] = triplet[1];
		    contributions[i][j].c[k] = FLT_MIN;
		    k++;
		    break;
		  }
		default:
		  {
		    /* cases FP_NORMAL and FP_ZERO */
		    contributions[i][j].x[k] = triplet[0];
		    contributions[i][j].y[k] = triplet[1];
		    contributions[i][j].c[k] = triplet[2];
		    k++;
		    break;
		  }
		}
	      n++;
	    }
	  contributions[i][j].nb_contribs = k;
	}
    }
  fclose (file_contributions);

  return (contributions);
}

/*===========================================================================*/
void
freeing_contributions (contributions_t ** MC, int nlig, int ncol)
{
  int i, j;

  for (i = 0; i < nlig; i++)
    {
      for (j = 0; j < ncol; j++)
	{
	  free (MC[i][j].x);
	  free (MC[i][j].y);
	  free (MC[i][j].c);
	}
      free (MC[i]);
    }
  free (MC);
  return;
}

/*===========================================================================*/
void
loading_matrices (char *filename,
		  long double mat_image_to_camera[4][4],
		  long double mat_camera_to_absolute[4][4])
{
  FILE *file_matrice = NULL;
  float mat_tmp[16];
  int i, j, k;
  int ret;

  file_matrice = fopen (filename, "rb");
  if (file_matrice == NULL)
    {
      perror ("loading_matrices\n");
      exit (EXIT_FAILURE);
    }

  if (mat_image_to_camera != NULL)
    {
      printf ("Matrix image to camera :\n");
      fseek (file_matrice, 2 * sizeof (int), SEEK_CUR);
      ret = fread (mat_tmp, sizeof (float), 16, file_matrice);
      if (ret != 16)
	{
	  perror ("read()\n");
	  exit (EXIT_FAILURE);
	}

      for (i = 0, k = 0; i < 4; i++)
	{
	  for (j = 0; j < 4; j++, k++)
	    {
	      printf ("%+10.5e ", mat_tmp[k]);
	      mat_image_to_camera[i][j] = (long double) mat_tmp[k];
	    }
	  printf ("\n");
	}
    }
  else
    {
      fseek (file_matrice, 2 * sizeof (int) + 16 * sizeof (float), SEEK_CUR);
    }


  if (mat_camera_to_absolute != NULL)
    {
      printf ("Matrix camera to absolute :\n");
      fseek (file_matrice, 2 * sizeof (int), SEEK_CUR);
      ret = fread (mat_tmp, sizeof (float), 16, file_matrice);
      if (ret != 16)
	{
	  perror ("read()\n");
	  exit (EXIT_FAILURE);
	}

      for (i = 0, k = 0; i < 4; i++)
	{
	  for (j = 0; j < 4; j++, k++)
	    {
	      printf ("%+10.5e ", mat_tmp[k]);
	      mat_camera_to_absolute[i][j] = (long double) mat_tmp[k];
	    }
	  printf ("\n");
	}
    }

  fclose (file_matrice);

  return;
}

/*===========================================================================*/
void
converting_coordinates (point_t ante_pt, point_t * neo_pt,
			long double M[4][4], int normer)
{
  int k, l;
  long double point[4];
  long double vec[4] = { 0.0, 0.0, 0.0, 0.0 };
  long double vx, vy, vz;

  point[0] = ante_pt.x;
  point[1] = ante_pt.y;
  point[2] = ante_pt.z;
  point[3] = ante_pt.w;

  /* matrix product transforming pixel into its representant into
   * the camera frame */
  for (k = 0; k < 4; k++)
    {
      for (l = 0; l < 4; l++)
	{
	  vec[k] += M[k][l] * point[l];
	}
    }

  /* normalization of the vector representing the direction of the
   * ray tracing */
  vx = vec[0];
  vy = vec[1];
  vz = vec[2];

  switch (normer)
    {
    case NORMALIZATION:
      {
	for (l = 0; l < 4; l++)
	  {
	    vec[l] /= sqrt (vx * vx + vy * vy + vz * vz);
	  }
	break;
      }
    case INV_NORMALIZATION:
      {
	for (l = 0; l < 4; l++)
	  {
	    vec[l] *= sqrt (vx * vx + vy * vy + vz * vz);
	  }
	break;
      }
    case NON_NORMALIZATION:
      {
	break;
      }

    }

  /* new components */
  neo_pt->x = vec[0];
  neo_pt->y = vec[1];
  neo_pt->z = vec[2];
  neo_pt->w = vec[3];
  return;
}

/*===========================================================================*/
void
writing_geometric_points (char *filename, contributions_t ** MC,
			  char *filename_matrices, int nlig, int ncol)
{
  int i, j, n;
  float distance;

  long double mat_image_to_camera[4][4];
  long double mat_camera_to_absolute[4][4];
  point_t pt_image, pt_camera, pt_absolute;

  int npix = nlig * ncol;
  int N;
  /* matrices of transformation */
  loading_matrices (filename_matrices, mat_image_to_camera,
		    mat_camera_to_absolute);

  /* file CSV */
  FILE *fichier = NULL;
  fichier = fopen (filename, "w");
  if (fichier == NULL)
    {
      perror ("writing_geometric_points\n");
      exit (EXIT_FAILURE);
    }

  /* header of the CSV file */
  fprintf (fichier, "X_geo_%d,Y_geo_%d,Z_geo_%d, X_pix_%d, Y_pix_%d\n",
	   ncol, nlig, npix, ncol, nlig);

  for (i = 0; i < nlig; i++)
    {
      for (j = 0; j < ncol; j++)
	{
	  N = MC[i][j].nb_contribs;

	  for (n = 0; n < N; n++)
	    {
	      /* computation of the geometric point P into the camera frame,
	         associated to pixel p
	       */
	      pt_image.x = MC[i][j].x[n];
	      pt_image.y = MC[i][j].y[n];
	      pt_image.z = 0;
	      pt_image.w = 1;

	      converting_coordinates (pt_image, &pt_camera,
				      mat_image_to_camera, NORMALIZATION);

	      distance = MC[i][j].c[n];

	      if (distance != FLT_MAX)
		{
		  pt_camera.x *= distance;
		  pt_camera.y *= distance;
		  pt_camera.z *= distance;
		  pt_camera.w = 1;

		  /* computation of the geometric point P into the absolute frame */
		  converting_coordinates (pt_camera, &pt_absolute,
					  mat_camera_to_absolute,
					  NON_NORMALIZATION);

		  if (isnormal (pt_absolute.x) && isnormal (pt_absolute.y)
		      && isnormal (pt_absolute.x))
		    {
		      fprintf (fichier, "%1.5Le, %1.5Le, %1.5Le, %d, %d\n",
			       pt_absolute.x, pt_absolute.y, pt_absolute.z, j,
			       i);
		    }
		}
	    }
	}
    }

  fclose (fichier);
  return;
}

/*===========================================================================*/
void
computing_geometric_points (char *filename_contributions,
			    char *filename_matrices, char *outfile)
{
  int nlig, ncol, ncbs;
  contributions_t **MC = NULL;

  int xmin = 0, ymin = 0, xmax = 0, ymax = 0;

  if (xmin == ymin && ymin == xmax && xmax == ymax && ymax == 0)
    {
      reading_image_dimensions (filename_contributions, &xmax, &ymax);
    }
  nlig = ymax - ymin + 1;
  ncol = xmax - xmin + 1;

  MC =
    loading_contributions (filename_contributions, xmin, ymin,
			   xmin + ncol - 1, ymin + nlig - 1, &ncbs);
  printf ("[OK] loading contributions\n");

  writing_geometric_points (outfile, MC, filename_matrices, nlig, ncol);

  freeing_contributions (MC, nlig, ncol);
  printf ("[INFO] end of processing\n");

  return;
}

/*===========================================================================*/
int
main (int argc, char **argv)
{
  if (argc == 1)
    {
      printf ("Usage : %s ", argv[0]);
      printf ("--help\n");
      exit (EXIT_SUCCESS);
    }
  else if (argc == 2 && !strcmp ("--help", argv[1]))
    {
      printf ("Usage : %s ", argv[0]);
      printf ("--help\n");
      printf ("Usage : %s ", argv[0]);
      printf ("--version\n");
      printf ("Usage : %s ", argv[0]);
      printf ("--infile file.cbs --matrix file.mat --outfile myfile.csv\n");
      exit (EXIT_SUCCESS);
    }
  else if (argc == 2 && !strcmp ("--version", argv[1]))
    {
      printf ("%s\nVersion %s\n", argv[0], VERSION);
      exit (EXIT_SUCCESS);
    }
  else if (argc == 7
	   && !strcmp ("--infile", argv[1])
	   && !strcmp ("--matrix", argv[3]) && !strcmp ("--outfile", argv[5]))
    {
      computing_geometric_points (argv[2], argv[4], argv[6]);
    }
  else
    {
      printf ("Usage : %s ", argv[0]);
      printf ("--help\n");
      exit (EXIT_SUCCESS);
    }

  exit (EXIT_SUCCESS);
}

/*===========================================================================*/
